use crate::config::{Config, OIDCProvider, OIDCProviderState};
use actix_session::Session;
use actix_web::{
    cookie::Cookie,
    get,
    http::{
        header::{ContentType, LOCATION},
        uri::PathAndQuery,
        StatusCode, Uri,
    },
    web::{self, ServiceConfig},
    HttpResponse, HttpResponseBuilder, Responder,
};
use openidconnect::{
    core::CoreAuthenticationFlow, http::HeaderValue, reqwest::async_http_client, url::Url,
    AuthorizationCode, CsrfToken, EndUserEmail, Nonce, RedirectUrl, TokenResponse,
};
use serde::{Deserialize, Serialize};
use std::{borrow::Cow, fmt::Write, marker::PhantomData};
use tinytemplate::TinyTemplate;

pub trait DynTemplate {
    fn name(&self) -> &'static str;
    fn text(&self) -> &'static str;
}

pub trait Template: Serialize
where
    PhantomData<Self>: DynTemplate,
{
    const NAME: &'static str;
    const TEXT: &'static str;
    fn render(&self) -> tinytemplate::error::Result<String> {
        with_templates(|templates| templates.render(Self::NAME, &self))
    }
}

impl<T: Template> DynTemplate for PhantomData<T> {
    fn name(&self) -> &'static str {
        T::NAME
    }

    fn text(&self) -> &'static str {
        T::TEXT
    }
}

macro_rules! templates {
    (
        $(
            #[text = $text:literal]
            $(#[$meta:meta])*
            $vis:vis struct $name:ident $fields:tt
        )*
    ) => {
        $(
            $(#[$meta])*
            $vis struct $name $fields

            impl Template for $name {
                const NAME: &'static str = stringify!($name);
                const TEXT: &'static str = $text;
            }
        )*
        const TEMPLATES: &'static [&'static dyn DynTemplate] = &[$(&PhantomData::<$name>,)*];
    };
}

#[derive(Debug, Serialize)]
pub struct SubscriptionLoggedOutTemplateProviders {
    pub pretty_name: String,
    pub url: String,
}

templates! {
    #[text = r#"<head><title>Subscription</title></head>
<body>
<ul>
{{- for provider in providers -}}
<li><a href="{provider.url}">{provider.pretty_name}</a></li>
{{- endfor -}}
</ul>
</body>"#]
    #[derive(Debug, Serialize)]
    pub struct SubscriptionLoggedOutTemplate {
        pub providers: Vec<SubscriptionLoggedOutTemplateProviders>,
    }
}

pub fn make_templates() -> TinyTemplate<'static> {
    let mut templates = TinyTemplate::new();
    for template in TEMPLATES {
        if let Err(e) = templates.add_template(template.name(), template.text()) {
            panic!("error parsing template {}: {}", template.name(), e);
        }
    }
    templates
}

pub fn with_templates<F: FnOnce(&TinyTemplate<'static>) -> R, R>(f: F) -> R {
    thread_local! {
        static TEMPLATES: TinyTemplate<'static> = make_templates();
    }
    TEMPLATES.with(f)
}

#[derive(Debug, Clone, Deserialize, Serialize)]
pub struct SessionState {
    pub email: EndUserEmail,
}

impl SessionState {
    pub fn get(session: &Session) -> Option<Self> {
        session
            .get("state")
            .expect("state is signed so we know it's valid")
    }
    pub fn set(&self, session: &Session) {
        session
            .insert("state", self)
            .expect("serialization can't fail");
    }
}

impl OIDCProvider {}

const NONCE: &'static str = "nonce";
const CSRF_TOKEN: &'static str = "csrf_token";

#[derive(Deserialize)]
struct AuthQuery {
    code: AuthorizationCode,
    state: CsrfToken,
}

#[get("/subscription/callback/{provider}")]
pub async fn callback(
    config: web::Data<Config>,
    session: Session,
    path: web::Path<(String,)>,
    query: web::Query<AuthQuery>,
) -> impl Responder {
    let mut resp = HttpResponse::SeeOther().body("");
    resp.headers_mut()
        .insert(LOCATION, "/subscription".parse().unwrap());
    if SessionState::get(&session).is_some() {
        return resp;
    }
    let Ok(Some(csrf_token)) = session.get::<CsrfToken>(CSRF_TOKEN) else {
        return resp;
    };
    let Ok(Some(nonce)) = session.get::<Nonce>(NONCE) else {
        return resp;
    };
    session.remove(CSRF_TOKEN);
    session.remove(NONCE);
    if csrf_token.secret() != query.state.secret() {
        return resp;
    }
    let Some(provider) = config.oidc.get(&path.0) else {
        return resp;
    };
    let token = match provider
        .state()
        .client
        .exchange_code(query.into_inner().code)
        .request_async(async_http_client)
        .await
    {
        Ok(token) => token,
        Err(e) => {
            log::warn!("Error getting token: {e}");
            return resp;
        }
    };
    dbg!(&token);
    let claims = match token
        .id_token()
        .unwrap()
        .claims(&provider.state().client.id_token_verifier(), &nonce)
    {
        Ok(claims) => claims,
        Err(e) => {
            log::warn!("Error verifying token: {e}");
            return resp;
        }
    };
    let Some(email) = claims.email().cloned() else {
        return resp;
    };
    SessionState { email }.set(&session);
    resp
}

#[get("/subscription/login/{provider}")]
pub async fn login(
    config: web::Data<Config>,
    session: Session,
    path: web::Path<(String,)>,
) -> impl Responder {
    let mut resp = HttpResponse::SeeOther().body("");
    resp.headers_mut()
        .insert(LOCATION, "/subscription".parse().unwrap());
    if SessionState::get(&session).is_some() {
        return resp;
    }
    let Some(provider) = config.oidc.get(&path.0) else {
        return resp;
    };
    let (url, csrf_token, nonce) = provider
        .state()
        .client
        .authorize_url(
            CoreAuthenticationFlow::AuthorizationCode,
            CsrfToken::new_random,
            Nonce::new_random,
        )
        .add_scopes(provider.scopes.clone())
        .url();
    session.insert(CSRF_TOKEN, csrf_token).unwrap();
    session.insert(NONCE, nonce).unwrap();
    resp.headers_mut()
        .insert(LOCATION, url.as_str().parse().unwrap());
    resp
}

#[get("/subscription")]
pub async fn subscription(config: web::Data<Config>, session: Session) -> impl Responder {
    if let Some(SessionState { email }) = SessionState::get(&session) {
        todo!()
    } else {
        let mut template = SubscriptionLoggedOutTemplate { providers: vec![] };
        for (name, provider) in config.oidc.iter() {
            template
                .providers
                .push(SubscriptionLoggedOutTemplateProviders {
                    pretty_name: provider.pretty_name.clone(),
                    url: format!("/subscription/login/{name}"),
                });
        }
        HttpResponse::Ok()
            .content_type(ContentType::html())
            .body(template.render().expect("rendering can't fail"))
    }
}

pub fn all_services(cfg: &mut ServiceConfig) {
    cfg.service(subscription).service(login).service(callback);
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_make_templates() {
        make_templates();
    }
}
